/* ====================================================================
 * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * Apache mod_rexx module.
 *
 * Author:  Carsten Mjartan
 * email:   carsten@mjartan.de
 */

/*

MODULE-DEFINITION-START
Name: rexx_module
ConfigStart
    LIBS="$LIBS -lregina -ldl"
    echo " + using -lregina -ldl libraries for mod_rexx"
ConfigEnd
MODULE-DEFINITION-END
*/


#define REGINA_REXX
/* #define OBJ_REXX */


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

/* Apache specific includes */

#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_main.h"
#include "http_log.h"
#include "util_script.h"
#include "http_conf_globals.h"
#include "multithread.h"
#include "util_script.h"

/* Rexx specific includes */

#define INCL_RXSYSEXIT
#define INCL_RXSHV


#ifdef OBJ_REXX
#include <rexx.h>                       /* Object Rexx: */
#else

#ifdef REGINA_REXX
#include <rexxsaa.h>                    /* Regina Rexx */
#else
#include <rexxtrans.h>                  /* rxtrans */
#endif

#endif


module MODULE_VAR_EXPORT rexx_module;



#define rexx_error(server,msg)  ap_log_error(APLOG_MARK,\
                                    APLOG_NOERRNO|APLOG_ERR,\
                                    server, msg)

#define rexx_debug(server,msg)  ap_log_error(APLOG_MARK,\
                                    APLOG_NOERRNO|APLOG_DEBUG,\
                                    server, msg)

/* Declaration of Handlers */

static void rexx_initialize(server_rec *s, pool *p);

static char *rexx_create_dir_config(pool *p, char *path);
static const char *rexx_cmd_createEnvironment(cmd_parms *parms, void *mconfig, char *yesno);

static int rexx_handler(request_rec *r);



/* Rexx-spezifische Kommandos in der Konfiguationsdatei */

static const command_rec rexx_commands[] =
{
    {   "RexxCreateEnvironment",
        rexx_cmd_createEnvironment,
        NULL,
        ACCESS_CONF,
        TAKE1,
        "yes/no as parameter, default is yes"
    },
    { NULL }
};

/* Content Handler nach MIME-Typ */

static const handler_rec rexx_handlers[] = {
    { "application/x-httpd-rexx",   rexx_handler },
    { "rexx-handler",               rexx_handler },
    { NULL }
};


/* module-structure */

module MODULE_VAR_EXPORT rexx_module =
{
    STANDARD_MODULE_STUFF,
    rexx_initialize,                /* module initializer */
    rexx_create_dir_config,         /* per-directory config creator */
    NULL,                           /* dir config merger */
    NULL,                           /* server config creator */
    NULL,                           /* server config merger */
    rexx_commands,                  /* command table */
    rexx_handlers,                  /* [9] list of handlers */
    NULL,                           /* [2] filename-to-URI translation */
    NULL,                           /* [5] check/validate user_id */
    NULL,                           /* [6] check user_id is valid *here* */
    NULL,                           /* [4] check access by host address */
    NULL,                           /* [7] MIME type checker/setter */
    NULL,                           /* [8] fixups */
    NULL,                           /* [10] logger */
#if MODULE_MAGIC_NUMBER >= 19970103
    NULL,                           /* [3] header parser */
#endif
#if MODULE_MAGIC_NUMBER >= 19970719
    NULL,                           /* process initializer */
#endif
#if MODULE_MAGIC_NUMBER >= 19970728
    NULL,                           /* process exit/cleanup */
#endif
#if MODULE_MAGIC_NUMBER >= 19970902
    NULL                            /* [1] post read_request handling */
#endif
};



LONG APIENTRY rxini_exit_handler(LONG ExitNumber, LONG Subfunction, PEXIT ParmBlock);
LONG APIENTRY rxsio_exit_handler(LONG ExitNumber, LONG Subfunction, PEXIT ParmBlock);


/* read-only: */
RXSYSEXIT exit_list[] = {
    {"rxini_handler", RXINI},
    {"rxsio_handler", RXSIO},
    {NULL, RXENDLST}
};


static mutex *rexx_mutex = NULL;




typedef struct {
    int createEnvironment;
} rexx_dir_config;



#ifndef WIN32

/* itoa converts integer values into a string        */
/* this function exists at VC 6.0 but not in linux   */

const char *itoa (int zahl, char *str, int delim)
{
    static const char ch[] = "0123456789ABCDEF";
    int z;
    int len = 0;
    int neg = 0;
    char *ptr = str;

    if ((delim < 2) || (delim > 16)) {
        fprintf(stderr,"mod_rexx: Error in itoa - invalid base value");
    exit(1);
    }

    if (zahl==0)
        return strcpy(str, "0");

    if (zahl < 0) { neg=1;  len++;  zahl=-zahl; }
    z = zahl;
    while (z > 0) { len++;  z /= delim; }
    ptr = str+len;
    *ptr = '\0';
    ptr--;

    while (zahl > 0) {
        *ptr = ch[zahl % delim];
        ptr--;
        zahl /= delim;
    }
    if (neg) *ptr='-';
    return str;
}
#endif

/* htoi converts hexadecimal strings into integer values for URL decoding */

static int htoi(char *s)
{
    int value;
    int c;

    c = s[0];
    if (isupper(c))
        c = tolower(c);
    value = (c >= '0' && c <= '9' ? c - '0' : c - 'a' + 10) * 16;

    c = s[1];
    if (isupper(c))
        c = tolower(c);
    value += c >= '0' && c <= '9' ? c - '0' : c - 'a' + 10;

    return (value);
}

/* URL-decoding: hexadecimal to ASCII - Conversion   */
/* ASCII-'+' to space conversion                     */

int url_decode(char *str, int len)
{
    char *dest = str;
    char *data = str;

    while (len--) {
        if (*data == '+')
            *dest = ' ';
        else if (*data == '%' && len >= 2 && isxdigit((int) *(data + 1)) && isxdigit((int) *(data + 2))) {
            *dest = (char) htoi(data + 1);
            data += 2;
            len -= 2;
        } else
            *dest = *data;
        data++;
        dest++;
    }
    *dest = '\0';
    return dest - str;
}


static void convert_to_rexx_var(char *str)
{
    char c;

    while (c = *str) {
        if ((c >= 'A' && c <= 'Z') || (c >= '0' && c <= '9'))
            ;
        else if (c >= 'a' && c <= 'z')
            *str += 'A'-'a';
        else if (c != '_' && c != '.' && c != '?' && c != '!')
            *str = '_';
        str++;
    }
}


/************************** Rexx Variable List functions  *************************/

typedef struct {
        PSHVBLOCK   first, last;
        char        *prefix;
        int         prefixlen;
        pool        *p;
} RexxVarList;

typedef RexxVarList*    PRexxVarList;


PRexxVarList rvl_create(pool *p, char *prefix) {
    PRexxVarList rvl = ap_palloc(p, sizeof(RexxVarList));

    rvl->first = rvl->last = NULL;
    rvl->prefix = prefix;
    rvl->prefixlen = strlen(prefix);
    rvl->p = p;
    return rvl;
}


PSHVBLOCK rvl_make_elem(PRexxVarList rvl, char *key, const char *val, int vallen) {
    PSHVBLOCK tmp = ap_palloc(rvl->p, sizeof(SHVBLOCK));
    int keylen;

    if (!tmp || !key) return NULL;

    keylen = strlen(key);
    if (keylen == 0) return NULL;
    if (vallen < 0) vallen = strlen(val);

    MAKERXSTRING(tmp->shvname, ap_pstrcat(rvl->p, rvl->prefix, key, NULL),
                                                    rvl->prefixlen + keylen);
    MAKERXSTRING(tmp->shvvalue, val, vallen);
    tmp->shvnamelen = rvl->prefixlen + keylen;
    tmp->shvvaluelen = vallen;
    tmp->shvcode = RXSHV_SET;
    tmp->shvret = RXSHV_OK;
    tmp->shvnext = NULL;

    return tmp;
}


PRexxVarList rvl_append_var(PRexxVarList rvl, char *key, const char *val, int vallen) {
    PSHVBLOCK tmp = rvl_make_elem (rvl, key, val, vallen);

    if (!tmp) return NULL;

    if (rvl->last) {
        rvl->last->shvnext = tmp;
        rvl->last = tmp;
    } else {
        rvl->first = rvl->last = tmp;
    }
    return rvl;
}


PRexxVarList rvl_append_arrvar(PRexxVarList rvl, char *key, char *val, int vallen) {
    PSHVBLOCK tmp, tmp2, new;
    int keylen;
    int count = 0;
    char ctbuf[20];

    if ((key == NULL) || (*key == '\0')) return NULL;
    keylen = strlen(key);

    if (rvl->first) {
        tmp = rvl->first;
        while ( tmp &&
                strncmp((tmp->shvname.strptr + rvl->prefixlen), key, keylen) ) {
            tmp = tmp->shvnext;
        }
        if (!tmp) {
            if (!rvl_append_var(rvl, ap_pstrcat(rvl->p, key , "1", NULL), val, -1) ||
                   !rvl_append_var(rvl, ap_pstrcat(rvl->p, key , "0", NULL), "1", 1)) {
                return NULL;
            }
        } else {
            while (tmp->shvnext && !strncmp((tmp->shvnext->shvname.strptr +
                                                rvl->prefixlen), key, keylen) ) {
                tmp2 = tmp;
                tmp = tmp->shvnext;
                count++;
            }
            /* tmp2 = last variable in 'key'-Array
               tmp  = 'key'-array counter */

            count++;
            itoa (count, ctbuf, 10);

            new = rvl_make_elem(rvl, ap_pstrcat(rvl->p, key, ctbuf, NULL), val, vallen);
            tmp->shvvalue.strptr = ap_pstrdup(rvl->p, ctbuf);
            tmp->shvvalue.strlength = tmp->shvvaluelen = strlen(ctbuf);
            tmp2->shvnext = new;
            new->shvnext = tmp;
            if (!new || !tmp->shvvalue.strptr) return NULL;
        }
    } else {
        if (!rvl_append_var(rvl, ap_pstrcat(rvl->p, key , "1", NULL), val, -1) ||
                rvl_append_var(rvl, ap_pstrcat(rvl->p, key , "0", NULL), "1", 1)) {
            return NULL;
        }
    }
    return rvl;
}


PRexxVarList rvl_merge(PRexxVarList list1, PRexxVarList list2) {
    if (!list1 || !list2) return NULL;

    if (!list1->first) {
        list1->first = list2->first;
        list1->last = list2->last;
    } else {
        if (list2->first) {
            list1->last->shvnext = list2->first;
            list1->last = list2->last;
        }
    }
    list1->prefix = "";
    list1->prefixlen = 0;

    return list1;
}

/***************************** POST data processing ****************************/


static int read_request_body (request_rec *r, char **rbuf, int *length)
{
    int rc;                 /* return code */

    long rlen;              /* number of bytes to read                 */
    int chunk_size;         /* maximal Size of the next chunk          */
    int len_read;           /* really read bytes per chunk             */

    char *ptr;              /* write position within the buffer        */


    /* The request body has to be transferred the classic way => in one piece      */
    if (rc = ap_setup_client_block (r, REQUEST_CHUNKED_ERROR) != OK) {
        rexx_error ( r->server, "Could not read request body");
        *rbuf = NULL;
        return rc;
    }

    if (ap_should_client_block(r)) {

        rlen = r->remaining;
        *length = rlen;

        /* reserve enough memory */
        *rbuf = (char *) ap_palloc(r->pool, rlen + 1 );

        if (!rbuf) {
            rexx_error (r->server, "Not enough memory to read request body");
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        ptr = *rbuf;
        chunk_size = rlen < HUGE_STRING_LEN ? rlen : HUGE_STRING_LEN;

        ap_hard_timeout ("read_request_body", r);

        while ( (len_read = ap_get_client_block(r, ptr, chunk_size)) > 0) {
            ap_reset_timeout(r);

            rlen -= len_read;
            ptr += len_read;

            chunk_size = rlen < HUGE_STRING_LEN ? rlen : HUGE_STRING_LEN;
        }
        ap_kill_timeout(r);
        *ptr = '\0';
    }
    return rc;
}




static int append_form_table(PRexxVarList rvl, const char *data)
{
    char *key, *val;
    int rc = OK;

    if (!rvl) return 0;

    while (*data && (val = ap_getword(rvl->p, &data, '&'))) {
        key = ap_getword(rvl->p, (const char **) &val, '=');

        ap_unescape_url((char *) key);
        convert_to_rexx_var(key);
        url_decode(val, strlen(val));

        if (key[strlen(key)-1] == '.')
            rvl_append_arrvar (rvl, key, val, -1);
        else
            rvl_append_var (rvl, key, val, -1);
    }
    return rc;
}


/************************ Rexx environment / header creation ***********************/

typedef struct {
    PRexxVarList rexx_vars;
    int     counter;
 /*   table   *rexx_vars; */
} header_data;


int append_header_var(void *v, const char *key, const char *val)
{
    header_data *hd = (header_data *) v;
    char ctbuf[20];
    char *keyname, *valname;

    hd->counter++;
    itoa(hd->counter, ctbuf, 10);

    keyname = ap_pstrcat(hd->rexx_vars->p, ctbuf, ".NAME", NULL);
    valname = ap_pstrcat(hd->rexx_vars->p, ctbuf, ".VALUE", NULL);

    rvl_append_var (hd->rexx_vars, keyname, key, -1);
    rvl_append_var (hd->rexx_vars, valname, val, -1);
    return 1;
}


/************************** process environment creation ****************************/

int request_setenv(void *v, const char *key, const char *val)
{
    char *env = malloc(strlen(key) + 1 + strlen(val) + 1);

    strcpy(env, key);
    strcat(env, "=");
    strcat(env, val);

    /* the microsoft msdn says that the memory allocated for env must not be freed */
    /* this may be a memory leak on other platforms (no docs)                      */
    putenv(env);
    return 1;
}


/************************ Rexx initialization / configuration ***********************/

static void rexx_initialize(server_rec *s, pool *p)
{
    if (!rexx_mutex)
        rexx_mutex = ap_create_mutex(NULL);
}


static char *rexx_create_dir_config(pool *p, char *path)
{
    rexx_dir_config *cfg = (rexx_dir_config *) ap_palloc(p, sizeof(rexx_dir_config));
    cfg->createEnvironment = 0;
    return (void *) cfg;
}


static const char *rexx_cmd_createEnvironment(cmd_parms *parms, void *mconfig, char *yesno)
{
    rexx_dir_config *cfg = (rexx_dir_config *) mconfig;

    if ( (!strcasecmp(yesno, "yes")) || (!strcasecmp(yesno, "on")) ) {
        cfg->createEnvironment = 1;
    } else if ( (!strcasecmp(yesno, "no")) || (!strcasecmp(yesno, "off")) ) {
        cfg->createEnvironment = 0;
    } else {
        return "parameter yes,no,on or off allowed";
    }
	return NULL;
}


/****************************** Rexx response handler ***************************/

typedef struct {
    request_rec     *request;
    int             headers_sent;
    PRexxVarList    rexx_vars;
    char            *rbuf, *rptr;
    int             rlen;
} request_data;



static int rexx_handler(request_rec *r)
{
    int	    returnCode;
    short   rc;                         /* RexxStart returnCodes      */
    const char	*content_type;          /* Content-type for POST-Body */

    rexx_dir_config *config;            /* per dir configuration      */
    request_data    *handler_info[2];
    request_data    *rdata;
    RXSTRING        result;             /* RexxStart result string    */

    header_data     *hd;
    char            ctbuf[20];

    PRexxVarList    rexx_glob, rexx_env, rexx_get, rexx_post;


    if (r->method_number != M_GET && r->method_number != M_POST) return DECLINED;
    if (r->finfo.st_mode == 0) return NOT_FOUND;

    rdata = (request_data *) ap_palloc (r->pool, sizeof(request_data));

    rdata->request = r;
    rdata->headers_sent = 0;
    rdata->rbuf = rdata->rptr = NULL;

    r->content_type = "text/html";

    ap_soft_timeout ("mod_rexx", r);

    if (r->header_only) {
        ap_send_http_header (r);
        rdata->headers_sent = 1;
        ap_kill_timeout(r);
        return OK;
    }

    rexx_glob   = rvl_create(r->pool, "");
    rexx_env    = rvl_create(r->pool, "APACHE!ENVIRONMENT.");
    rexx_get    = rvl_create(r->pool, "APACHE!QUERY.");
    rexx_post   = rvl_create(r->pool, "APACHE!FORM.");

    if (!rexx_glob || !rexx_env || !rexx_get || !rexx_post) {
        rexx_error (r->server, "Not enough memory to create Rexx variable lists");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    /* generate environment and create rexx-variables from it */
    ap_add_common_vars(r);
    ap_add_cgi_vars(r);

    /* append environment to rexx variables in APACHE.ENVIRONMENT.* */
    hd = (header_data *) ap_palloc(r->pool, sizeof(header_data));
    hd->counter = 0;
    hd->rexx_vars = rexx_env;

    ap_table_do (append_header_var, (void *) hd, r->subprocess_env, NULL);

    itoa(hd->counter, ctbuf, 10);
    rvl_append_var (rexx_env, "0", ap_pstrdup(r->pool, ctbuf), -1);


    /* QueryString?   */

    if (r->args) {
        append_form_table (rexx_get, r->args);
        rvl_append_var(rexx_glob, "APACHE!QUERY_STRING", r->args, -1);
    } else
        rvl_append_var(rexx_glob, "APACHE!QUERY_STRING", "", 0);


    /* POST request: */

    if (r->method_number == M_POST) {
        /* Daten einlesen, Abbruch, falls Fehler */

        if ((rc = read_request_body (r, &rdata->rbuf, &rdata->rlen)) != OK)
            return rc;

        rdata->rptr = rdata->rbuf;

        content_type = ap_table_get (r->headers_in, "Content-Type");
        if (strcasecmp(content_type, "application/x-www-form-urlencoded") == 0) {
            append_form_table(rexx_post, rdata->rbuf);
        }
        rvl_append_var (rexx_glob, "APACHE!POST_DATA", rdata->rbuf, rdata->rlen);
    } else {
        rvl_append_var (rexx_glob, "APACHE!POST_DATA", "", 0);
    }

    /* merge Rexx variable lists */

    rvl_merge (rexx_glob, rexx_env);
    rvl_merge (rexx_glob, rexx_get);
    rvl_merge (rexx_glob, rexx_post);
    rdata->rexx_vars = rexx_glob;

    /* prepare exit handlers */

    handler_info[0] = rdata;
    handler_info[1] = NULL;

    RexxRegisterExitExe ("rxini_handler", (PFN) &rxini_exit_handler, (PUCHAR) handler_info);
    RexxRegisterExitExe ("rxsio_handler", (PFN) &rxsio_exit_handler, (PUCHAR) handler_info);

    /*  In multithreading environments like Windows the next steps must not overlap
        in concurrent threads, because there is only one CGI environment available for
        all threads in one process.
    */
    config = (rexx_dir_config *) ap_get_module_config(r->per_dir_config, &rexx_module);

    if (config->createEnvironment) {
        ap_acquire_mutex(rexx_mutex);

        /* update environment */
        ap_table_do (request_setenv, (void *) r, r->subprocess_env, NULL);
    }


    result.strptr = NULL;         /* Rexx has to allocate memory for the result string */

    returnCode = RexxStart(0, NULL,
                    r->filename,
                    NULL,
                    "",
                    RXCOMMAND,
                    exit_list,
                    &rc,
                    &result);

    rdata->request = NULL;

    if (config->createEnvironment)
        ap_release_mutex(rexx_mutex);

    RexxDeregisterExit("rxini_handler", NULL);
    RexxDeregisterExit("rxsio_handler", NULL);

    if (result.strptr)
        free(result.strptr);

    ap_kill_timeout(r);
    return OK;
}



LONG APIENTRY rxini_exit_handler(LONG ExitNumber, LONG Subfunction, PEXIT ParmBlock)
{
    USHORT          query_flag;
    request_data    *handler_info[2];
    request_data    *rdata;

    if (ExitNumber != RXINI || Subfunction != RXINIEXT) return RXEXIT_NOT_HANDLED;

    /* get request information */
    RexxQueryExit("rxini_handler", NULL, &query_flag, (PUCHAR) handler_info);
    rdata = handler_info[0];

    /* give the variable list to Rexx */
    if (rdata->rexx_vars->first) {
        RexxVariablePool(rdata->rexx_vars->first);
    }

    return RXEXIT_HANDLED;
}



LONG APIENTRY rxsio_exit_handler(LONG ExitNumber, LONG Subfunction, PEXIT ParmBlock)
{
    USHORT          query_flag;
    request_data    *handler_info[2];
    request_data    *rdata;
    request_rec     *request;

    RXSTRING str;
    char    *hkey, *hval;

    RXSIOTRD_PARM   *trd;
    char    *t;
    int     len;


    if (ExitNumber != RXSIO) return RXEXIT_NOT_HANDLED;

    /* get request information */
    RexxQueryExit("rxsio_handler", NULL, &query_flag, (PUCHAR) handler_info);
    rdata = handler_info[0];
    request = rdata->request;

    if (!request) return RXEXIT_RAISE_ERROR;

    switch (Subfunction) {
        case RXSIOTRC:
            /* trace- or error messages are redirected */

            if (!rdata->headers_sent) {
                ap_send_http_header (request);
                rdata->headers_sent = 1;
            }

            str = ((RXSIOTRC_PARM *)ParmBlock)->rxsio_string;

            ap_rputs("</TD></TR></TABLE><B>", request);
            ap_rwrite(str.strptr, str.strlength, request);
            ap_rputs("</B><BR>\n", request);
            break;

        case RXSIOSAY:
            /* text output by SAY is redirected */

            str = ((RXSIOSAY_PARM *)ParmBlock)->rxsio_string;

            if (!rdata->headers_sent) {
                if (str.strlength) {
                    /* Zur Headerliste hinzufgen: */
                    hkey = ap_pstrndup(request->pool, str.strptr, str.strlength);
                    if (hval = strchr(hkey, ':')) {
                        *hval = '\0';
                        while (*(++hval) == ' ');
                        if (! strcmp(hkey, "Content-type")) {
                            request->content_type = hval;
                        } else {
                            ap_table_set (request->headers_out, hkey, hval);
                        }
                    }
                } else {
                    /* empty line ==> send headers */
                    ap_send_http_header (request);
                    rdata->headers_sent = 1;
                }
            } else {
                ap_rwrite(str.strptr, str.strlength, request);
                ap_rputc('\n', request);
            }
            break;

        case RXSIOTRD:
            /* redirection of PULL / PARSE PULL input */

            if (rdata->rbuf) {
                t = rdata->rptr;
                if (t < (rdata->rbuf + rdata->rlen)) {
                    while ((t < (rdata->rbuf + rdata->rlen)) && (*t != 13) && (*t != 10))
                        t++;
                    trd = (RXSIOTRD_PARM *) ParmBlock;
                    len = t - rdata->rptr;
                    trd->rxsiotrd_retc.strlength = len;
#ifdef OBJ_REXX
                    if (len > 256) {
                        if (! (trd->rxsiotrd_retc.strptr = (char *) malloc(len))) {
                            rexx_error ( request->server, "Could not allocate memory for Rexx Input");
                            return RXEXIT_RAISE_ERROR;
                        }
                    }
                    if (len) strncpy(trd->rxsiotrd_retc.strptr, rdata->rptr, len);
#else
                    trd->rxsiotrd_retc.strptr = rdata->rptr;
#endif
                    if ((t < (rdata->rbuf + rdata->rlen - 1)) && (*t==13) && (*(t+1)==10))
                        rdata->rptr = t+2;
                    else
                        rdata->rptr = t+1;
                } else
                    return RXEXIT_RAISE_ERROR;
            } else
                return RXEXIT_RAISE_ERROR;
            break;
        case RXSIODTR:
        default:
            return RXEXIT_NOT_HANDLED;
    }
    return RXEXIT_HANDLED;
}

